/**
 * This Cloud Function implements an DIY integration to the Pangolin platform
 *
 * @param {!Object} req Cloud Function request context.
 * @param {!Object} res Cloud Function response context.
 */


var config = {
    endpoint: 'http://PANGOLIN_URL.TLD', // Pangolin endpoint (The actual platform we want to integrate with)
    username: 'EXAMPLE_USER', // Basic auth username
    password: 'EXAMPLE_PASSWORD', // Basic auth password
    shared_secret: 'EXAMPLE_HMAC_SECRET' // HMAC shared secret for validating the payload
};

// Include required libraries
const request = require('request');
const Promise = require('promise');
const crypto = require('crypto');
const auth = require('basic-auth');

/**
 * This method gets called every time an event (HTTP request) is send to the Google Cloud function.
 * @param req The incoming HTTP request
 * @param res The HTTP response to return
 * @returns {*}
 */
exports.diy = function diy(req, res) {
    console.log('Incoming request: ', req);

    // Perform basic authentication
    var user = auth(req);
    if (typeof user === 'undefined' || user.name !== config.username || user.pass !== config.password) {
        res.statusCode = 401;
        res.setHeader('WWW-Authenticate', 'Basic realm="diy-integration"');
        res.end('Access denied');
        return;
    }

    // Perform HMAC payload validation
    if (typeof req.headers['x-hmac'] === 'undefined') {
        return res.status(403).json({error: 'Header x-hmac not defined'});
    }
    var hmac = crypto.createHmac('sha1', config.shared_secret);

    // Because Google Cloud functions automatically parses the JSON in the body for us we need to stringify it again in order to calculate the HMAC
    var hmacJsonBody = JSON.stringify(req.body);
    hmac.update(hmacJsonBody);
    var expectedHmac = hmac.digest('hex');
    if (expectedHmac !== req.headers['x-hmac']) {
        return res.status(403).json({error: 'Incorrect x-hmac expected: ' + expectedHmac + ' Based on the following body: ' + hmacJsonBody});
    }

    // Check that the required URL parameter type is defined
    if (typeof req.query.type === 'undefined') {
        return res.status(400).json({error: 'Param type is not defined'});
    }

    // Perform routing to the correct function based on the HTTP method and the URL parameter type
    var methodUpper = req.method.toUpperCase();
    if (methodUpper === 'POST') {
        switch (req.query.type) {
            case 'CREATE_AUDIENCE':
                // Call the Create Audience method
                createAudience(req.body.Name).then(function (response) {
                    return res.status(200).json({
                        id: response.created
                    });
                });
                break;
            case 'ADD_TO_AUDIENCE':
                // Call the add to audience method for each of the specified mutations
                var promises = [];
                for (var idx in req.body.Mutations) {
                    if (!req.body.Mutations.hasOwnProperty(idx)) {
                        continue;
                    }
                    var mutation = req.body.Mutations[idx];
                    promises.push(addUser(mutation.Audience, mutation.Uuid));
                }

                Promise.all(promises).then(function () {
                    return res.status(200).json({
                        success: true
                    });
                });
                break;
            default:
                return res.status(400).json({error: 'Invalid value for param type for method POST'});
        }
    } else if (methodUpper === 'GET') {
        switch (req.query.type) {
            case 'VALIDATE':
                return res.status(200).json({msg: 'Hello there, good to see you!'});
            default:
                return res.status(400).json({error: 'Invalid value for param type for method GET'});
        }
    } else {
        return res.status(400).json({error: 'Unsupported method'});
    }

    /**
     * Create an Audience in Pangolin and return a promise that will be resolved once the audience is created. The resolved promise will contain the parsed JSON that is
     * returned from Pangolin
     * @param name The name of the audience to create
     * @returns Promise
     */
    function createAudience(name) {
        return new Promise(function (resolve, reject) {
            // Forward the request to Pangolin
            request.post(config.endpoint + 'segment/create?name=' + encodeURIComponent(name), {}, function (e, r, body) {
                resolve(JSON.parse(body));
            });
        });
    }

    /**
     * Add a user to a Pangolin audience
     * @param audienceId The id of the audience in Pangolin to add the user to
     * @param userId The user id in Pangolin to add to the audience
     * @returns Promise
     */
    function addUser(audienceId, userId) {
        return new Promise(function (resolve, reject) {
            request.post(config.endpoint + 'segment/upload?segmentId=' + encodeURIComponent(audienceId) + '&userId=' + encodeURIComponent(userId), {}, function(e, r, body) {
                // Pangolin returns an empty response on success
                if (body.length === 0) {
                    resolve();
                } else {
                    // Try to parse the error response
                    var resp;
                    try {
                        resp = JSON.parse(body);
                    } catch (e) {
                        resp = e;
                    }
                    reject(resp);
                }
            });
        });
    }
};
